/*	Manage communication with Blocks for setting and subscribing to properties.

	Copyright(c) 2022 PIXILAB Technologies AB, Sweden (http://pixilab.se). All Rights Reserved.
 */

using System.Collections.Generic;
using System;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using UnityEngine;
using System.Collections.Concurrent;


/*	The following is used to decode JSON data, and assumes Packages/manifest.json contains
	"com.unity.nuget.newtonsoft-json": "2.0.0"
*/
using Newtonsoft.Json;

public delegate void ObjMsg(object msg);
delegate void StrMsg(string message);

class Event {
	private readonly StrMsg mToCall;
	private readonly string mDataToPass;

	public Event(StrMsg toCall, string data) {
		mToCall = toCall;
		mDataToPass = data;
	}

	public void Handle() {
		mToCall(mDataToPass);
	}
}

public class BlocksTalker {
	private TcpClient socket;
	private volatile bool closed = false;
	private Thread rceiverThread;
	private NetworkStream stream;
	private ConcurrentQueue<Event> events; // From listener thread to main thread
	private Queue<string> sendQueue;    // For deferred transmission if not yet ready
	private Dictionary<string, HashSet<ObjMsg>> subscriptions;

	private readonly string serverAddress;
	byte[] readBuf = new byte[4096]; // Buffer to accumulate data into

	/*	Create this talker, talking to the specified network address.
	 */
	public BlocksTalker(string talkTo = "localhost") {
		serverAddress = talkTo;
		events = new ConcurrentQueue<Event>();
		sendQueue = new Queue<string>();
		subscriptions = new Dictionary<string, HashSet<ObjMsg>>();
		Start();
	}

	/**
	 * Call regularly to maintain data communication.
	 */
	public void Maintain() {
		if (events.TryDequeue(out Event message))
			message.Handle();
	}

	/**
	 * Call on application shut down to close the connection.
	 */
	public void Close() {
		if (socket != null) {
			stream.Close();
			socket.Close();
			Debug.Log("Socket closed");
			closed = true;
		}
	}

	/**
	 * Send command to Blocks to set propertyPath to value, which should be a
	 * value of the same type as the property (i.e., string, number or boolean).
	 */
	public void SetProperty(string propertyPath, object value) {
		string valueStr;
		if (value is Boolean) // Make sure we get valid JSON values here
			valueStr = (bool)value ? "true" : "false";
		else if (value is String) // Quote the string
			valueStr = '"' + value.ToString() + '"';
		else
			valueStr = value.ToString();

		string toSend = "{ \"type\": \"prop\", \"path\": \"" + propertyPath +
			"\", \"value\": " + valueStr + "}\r";
		if (socket != null && socket.Connected)
			SendMessage(toSend);
		else
			putOnSendQueue(toSend);
	}

	private void putOnSendQueue(string toSend) {
		if (sendQueue.Count < 50)
			sendQueue.Enqueue(toSend);
		else
			Debug.Log("sendQueue overflow - data dropped");
	}

	public void SubscribeProperty(string propertyPath, ObjMsg dataHandler) {
		HashSet<ObjMsg> handlers;
		if (!subscriptions.TryGetValue(propertyPath, out handlers)) {
			handlers = new HashSet<ObjMsg>();
			subscriptions.Add(propertyPath, handlers);
			if (socket != null && socket.Connected)
				SendMessage(MakeSubscribePropertyMsg("sub", propertyPath) + '\r');
		}
		handlers.Add(dataHandler);
	}

	public void UnsubscribeProperty(string propertyPath, ObjMsg dataHandler) {
		HashSet<ObjMsg> handlers;
		if (subscriptions.TryGetValue(propertyPath, out handlers)) {
			if (handlers.Remove(dataHandler)) {
				if (handlers.Count == 0) {
					subscriptions.Remove(propertyPath);
					if (socket != null && socket.Connected)
						SendMessage(MakeSubscribePropertyMsg("unsub", propertyPath) + '\r');
				}
			} else
				Debug.Log("UnsubscribeProperty handler not subscribed: " + propertyPath);
		} else
			Debug.Log("UnsubscribeProperty not subscribed: " + propertyPath);
	}

	private string MakeSubscribePropertyMsg(string type, string propertyPath) {
		return "{ \"type\": \"" + type + "\", \"path\": \"" + propertyPath + "\"}";
	}

	/**
	 * Callback from receiver thread once we know we're connected OK.
	 * Send any queued messages and open subscriptions.
	 */
	private void BlocksConnected(string notUsed) {
		while (socket.Connected && sendQueue.Count > 0)
			SendMessage(sendQueue.Dequeue());
		if (subscriptions.Count > 0)
			SendSubscriptions();
	}

	private void SendSubscriptions() {
		// First collect subscriptions as a list of complete messages
		LinkedList<string> subMsgs = new LinkedList<string>();
		foreach (string path in subscriptions.Keys)
			subMsgs.AddLast(MakeSubscribePropertyMsg("sub", path));
		// Then send those in batches of reasonable size
		StringBuilder sb = new StringBuilder();
		foreach (string subMsg in subMsgs) {
			if (sb.Length == 0)
				sb.Append('[');
			else
				sb.Append(',');
			sb.Append(subMsg);

			if (sb.Length > 2048) {
				sb.Append("]\r");
				if (!socket.Connected)	// Bail if connection dropped meanwhile
					return;
				SendMessage(sb.ToString());
				sb.Clear();
			}
		}
		if (sb.Length > 0) {
			sb.Append("]\r");
			if (!socket.Connected)  // Bail if connection dropped
				return;
			SendMessage(sb.ToString());
		}
	}

	/**	Received property values update from Blocks as JSON. Notify subscribers.
	 */
	private void BlocksMsgReceived(string msg) {
		List<Dictionary<string, object>> props =
			JsonConvert.DeserializeObject<List<Dictionary<string, object>>>(msg);
		foreach (Dictionary<string, object> propDict in props) {
			if (propDict.TryGetValue("path", out object path) &&
				propDict.TryGetValue("value", out object value)
			) {
				NotifySubscriptions(path.ToString(), value);
			}
		}
	}

	/**	Notify all subsribers toProp withValue.
	 */
	private void NotifySubscriptions(string toPath, object withValue) {
		if (this.subscriptions.TryGetValue(toPath, out HashSet<ObjMsg> receivers)) {
			foreach (ObjMsg receiver in receivers)
				receiver(withValue);
		}
	}

	/**	Start up thread performing TCP connection and data receive.
	 */
	private void Start() {
		rceiverThread = new Thread(new ThreadStart(CommsThreadLoop)) {
			IsBackground = true
		};
		rceiverThread.Start();
		// Debug.Log("clientReceiveThread started");
	}

	/**
	 * Connection and data listener thread loop function.
	 */
	private void CommsThreadLoop() {
		bool toldConnFailed = false;
		bool toldConnected = false;

		try {
			while (!closed) {
				do {
					try {
						socket = new TcpClient(serverAddress, 3041);
					} catch (SocketException socketException) {
						if (!toldConnFailed) {
							toldConnFailed = true;
							Debug.Log("Failed connecting socket " + socketException.ToString());
						}
						Thread.Sleep(4000);	// Try again in a bit
					}
				} while (socket == null && !closed);
				if (socket.Connected && !toldConnected) {
					toldConnected = true;
					events.Enqueue(new Event(BlocksConnected, ""));
				}

				using (stream = socket.GetStream()) {
					try {
						ReadData();
					} catch (Exception ex) {
						Debug.Log("Exception reading data: " + ex.ToString());
					}
				}
				socket = null;  // Socket gone - need to obtain a new one
				toldConnected = false;
			}
		} catch (Exception socketException) {
			if (socket != null) // Skip this message if socket already closed
				Debug.Log("Unexpected socket exception: " + socketException.ToString());
		}
	}

	/**
	 * Blocking funciton to read data from socket's stream, look for CR framing character, 
	 * then package as string and enqueue for further processing on main thread.
	 */
	private void ReadData() {
		while (socket.Connected && !closed) {
			int readPos = 0;    // Current position to read data into in readBuf
			int bufSpace = readBuf.Length;
			int readBytes;      // Number of bytes read from stream each time
			int crPos;			// Position of CR framing char

			// Read raw data bytes from stream into readBuf
			while ((readBytes = stream.Read(readBuf, readPos, bufSpace)) > 0) {
				int bufBytes = readPos + readBytes; // Total number of bytes in buffer now
				readPos += readBytes;	// Position where to read next byte to unless changed below
				// Got some more data - look for data frames
				while (bufBytes > 0 && (crPos = Array.IndexOf(readBuf, (byte)0x0D, 0, bufBytes)) >= 0) {
					string serverMessage = Encoding.UTF8.GetString(readBuf, 0, crPos);
					events.Enqueue(new Event(BlocksMsgReceived, serverMessage));
					if (bufBytes > crPos + 1) { // Preserve trailing data
						int removedBytes = crPos + 1;
						Array.Copy(readBuf, removedBytes, readBuf, 0, bufBytes - removedBytes);
						readPos -= removedBytes;    // Acount for data removed from buf above
						bufBytes -= removedBytes;
					} else {  // Nothing to preserve - reset to start of buffer
						readPos = 0;
						bufBytes = 0;
					}
					bufSpace = readBuf.Length - readPos;
				}
			}
		}
	}

	/**
	 * Send message to Blocks.
	 */
	private void SendMessage(string message) {
		if (socket == null)
			Debug.Log("Socket not available");
		else {
			try {
				byte[] clientMessageAsByteArray = Encoding.UTF8.GetBytes(message);
				stream.Write(clientMessageAsByteArray, 0, clientMessageAsByteArray.Length);
				stream.Flush();
				if (!socket.Connected) // Apparently failed - need to try this again once connected
					putOnSendQueue(message);
			} catch (SocketException socketException) {
				Debug.Log("SocketException: " + socketException.ToString());
			}
		}
	}
}
